
/*
LoaderGenerator - tool for generated xml, sql and doml file needed for Octopus.


    Copyright (C) 2003  Together

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

package org.webdocwf.util.loader.generator;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.FactoryConfigurationError;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.DOMException;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import java.io.File;
import java.io.IOException;
import java.io.*;
import java.sql.*;


import org.enhydra.xml.*;
import org.webdocwf.util.loader.LoaderException;

/**
 * LoaderGenerator class dinamicly generates the xml files (LoaderJob.xml and ImportDefinition.xml)
 * from the input data and source database.
 * Also LoaderGenerator creates SQL statements for building and loading an SQL target database.
 * @author Radoslav Dutina
 * @version 1.0
 */
public class LoaderGenerator {

  private InputParameters generatorParameters;

  /**
   * Construct object LoaderGenerator with associated parameters.
   * @param sourceType defines the type of the source data.
   * @param targetType defines the type of the target data.
   * @param sourceDriverName is driver of defined source database.
   * @param targetDriverName is driver of defined  target database.
   * @param sourceDataBase defines the place where you're put the database tables.
   * @param targetDataBase defines the place where the target database tables are.
   * @param valueMode defines the difference from overwrite and update attribute.
   * Default valueMode is Overwrite.
   * @param generatorOutput is optional. That represents the directory, where the
   * Octopus place created files. If this argument don't exists, Octopus place
   * created files in to current directory!
   * @param sourceUser defines the user of the source database.
   * @param sourcePassword defines user password for the source database.
   * @param targetUser defines the user of the target database.
   * @param targetPassword defines user password for the target database.
   * @param domlPath defines if the doml file exists. Default value is false
   * all, application generates xml, sql and doml files as ouput files. If the generateDoc has value
   * doml, application generates only the doml file, and if generateDoc has the value wdoml, application
   * generates xml and sql files.
   * @param packageName defines the package name for the generated doml file.
   * @param generateDropTableStmt defines if you want to generate only Sql statement for DROP TABLE
   * @param generateCreateTableStmt defines if you want to generate only Sql statement for CREATE TABLE
   * @param generateCreatePKStmt defines if you want to generate only Sql statement for ALTER TABLE
   * @param generateCreateFKStmt defines if you want to generate only Sql statement for ALTER TABLE
   * @param generateCreateIndexStmt defines if you want to generate only Sql statement for ALTER TABLE
   * @param generateSqlForAllVendors defines if you want to generate Sql statements for all named
   * database vendors
   * @param generateXml defines if you want to generate xml file as output files
   * @param generateDoml defines if you want to generate doml file as output file
   * @param fullMode defines in which mode you want to generate xml files.
   * @param includeTableList defines the list of tables which you want to include into
   * Generator process
   * @param confJarStructure defines the structure of jar file in which are conf
   * files placed
   * @throws LoaderException
   */

  public LoaderGenerator(String sourceType, String sourceDataBase, String valueMode,
                         String generatorOutput, String sourceDriverName, String targetDriverName,
                         String targetDataBase, String targetType, String sourceUser,
                         String sourcePassword, String targetUser, String targetPassword,
                         String domlPath, String packageName,
                         String generateDropTableStmt,String generateDropIntegrityStmt, String generateCreateTableStmt,
                         String generateCreatePKStmt, String generateCreateFKStmt,
                         String generateCreateIndexStmt, String generateSqlForAllVendors,
                         String generateXml, String generateDoml, String fullMode,
                         String includeTableList, String confJarStructure) throws LoaderException{

    this(sourceType, sourceDataBase, valueMode,generatorOutput, sourceDriverName,
          targetDriverName,targetDataBase, targetType, sourceUser,
          sourcePassword, targetUser, targetPassword,domlPath, packageName,
          generateDropTableStmt,generateDropIntegrityStmt,generateCreateTableStmt,generateCreatePKStmt,                            generateCreateFKStmt,
          generateCreateIndexStmt, generateSqlForAllVendors,generateXml, generateDoml,
          fullMode, "false",includeTableList,confJarStructure);

  }


  /**
   * Construct object LoaderGenerator with associated parameters.
   * @param sourceType defines the type of the source data.
   * @param targetType defines the type of the target data.
   * @param sourceDriverName is driver of defined source database.
   * @param targetDriverName is driver of defined  target database.
   * @param sourceDataBase defines the place where you're put the database tables.
   * @param targetDataBase defines the place where the target database tables are.
   * @param valueMode defines the difference from overwrite and update attribute.
   * Default valueMode is Overwrite.
   * @param generatorOutput is optional. That represents the directory, where the
   * Octopus place created files. If this argument don't exists, Octopus place
   * created files in to current directory!
   * @param sourceUser defines the user of the source database.
   * @param sourcePassword defines user password for the source database.
   * @param targetUser defines the user of the target database.
   * @param targetPassword defines user password for the target database.
   * @param domlPath defines if the doml file exists. Default value is false
   * all, application generates xml, sql and doml files as ouput files. If the generateDoc has value
   * doml, application generates only the doml file, and if generateDoc has the value wdoml, application
   * generates xml and sql files.
   * @param packageName defines the package name for the generated doml file.
   * @param generateDropTableStmt defines if you want to generate only Sql statement for DROP TABLE
   * @param generateCreateTableStmt defines if you want to generate only Sql statement for CREATE TABLE
   * @param generateCreatePKStmt defines if you want to generate only Sql statement for ALTER TABLE
   * @param generateCreateFKStmt defines if you want to generate only Sql statement for ALTER TABLE
   * @param generateCreateIndexStmt defines if you want to generate only Sql statement for ALTER TABLE
   * @param generateSqlForAllVendors defines if you want to generate Sql statements for all named
   * database vendors
   * @param generateXml defines if you want to generate xml file as output files
   * @param generateDoml defines if you want to generate doml file as output file
   * @param fullMode defines in which mode you want to generate xml files.
   * @param restoreMode defines if we use application for restoring the database
   * @param includeTableList defines the list of tables which you want to include into
   * Generator process
   * @param confJarStructure defines the structure of jar file in which are conf files
   * placed.
   * @throws LoaderException
   */

  public LoaderGenerator(String sourceType, String sourceDataBase, String valueMode,
                         String generatorOutput, String sourceDriverName, String targetDriverName,
                         String targetDataBase, String targetType, String sourceUser,
                         String sourcePassword, String targetUser, String targetPassword,
                         String domlPath, String packageName,
                         String generateDropTableStmt,String generateDropIntegrityStmt ,String generateCreateTableStmt,
                         String generateCreatePKStmt, String generateCreateFKStmt,
                         String generateCreateIndexStmt, String generateSqlForAllVendors,
                         String generateXml, String generateDoml, String fullMode,
                         String restoreMode, String includeTableList,
                         String confJarStructure) throws LoaderException {

    generatorParameters=new InputParameters(confJarStructure);
    generatorParameters.setConfJarStructure(confJarStructure);

    generatorParameters.setSourceType(sourceType);
    //target and source driver name must goes frist, because alterTablePrimaryKey...
    generatorParameters.setTargetDriverName(targetDriverName);
    generatorParameters.setSourceDriverName(sourceDriverName);
    generatorParameters.setTargetType(targetType);
    generatorParameters.setSourceDataBase(sourceDataBase);
    generatorParameters.setTargetDataBase(targetDataBase);
    generatorParameters.setValueMode(valueMode);
    generatorParameters.setGeneratorOutput(generatorOutput);
    generatorParameters.setSourceUser(sourceUser);
    generatorParameters.setSourcePassword(sourcePassword);
    generatorParameters.setTargetUser(targetUser);
    generatorParameters.setTargetPassword(targetPassword);

    if(generateDropTableStmt!=null)
      generatorParameters.setSqlStmtDropTable(generateDropTableStmt);
    if(generateDropIntegrityStmt!=null)
      generatorParameters.setSqlStmtDropIntegrity(generateDropIntegrityStmt);
    if(generateCreateTableStmt!=null)
      generatorParameters.setSqlStmtCreateTable(generateCreateTableStmt);
    if(generateCreatePKStmt!=null)
      generatorParameters.setSqlStmtCreatePK(generateCreatePKStmt);
    if(generateCreateFKStmt!=null)
      generatorParameters.setSqlStmtCreateFK(generateCreateFKStmt);
    if(generateCreateIndexStmt!=null)
      generatorParameters.setSqlStmtCreateIndex(generateCreateIndexStmt);
    if(generateSqlForAllVendors!=null)
      generatorParameters.setAllVendors(generateSqlForAllVendors);

    generatorParameters.setDomlPath(domlPath);
    if(generateXml!=null)
      generatorParameters.setGenerateXml(generateXml);
    if(generateDoml!=null)
      generatorParameters.setGenerateDoml(generateDoml);
    generatorParameters.setPackageName(packageName);
    if(fullMode!=null)
      generatorParameters.setFullMode(fullMode);
    if(restoreMode!=null)
      generatorParameters.setRestoreMode(restoreMode);
    if(includeTableList!=null)
      generatorParameters.setIncludeTableList(includeTableList);

      generatorParameters.setSourceDriverProperties();
      generatorParameters.setTargetDriverProperties();
      generatorParameters.setOidAndVersionColumnName();

  }

  /**
   * Public constructor of LoaderGenerator class. Constructor set value for sourceDataBase and
   * targetDataBase attribute. Class set all other attributes to the default values.
   * @param sourceDataBase defines the place where you're put the database tables.
   * @param targetDataBase defines the place where the target database tables are.
   * @param sourceType defines the type of the source database.
   * @param targetType defines the type of the target database.
   * @throws LoaderException
   */
  public LoaderGenerator(String sourceDataBase, String targetDataBase, String sourceType,
                         String targetType) throws LoaderException {

    this(sourceDataBase,targetDataBase,sourceType,targetType,null);

  }

  /**
 * Public constructor of LoaderGenerator class. Constructor set value for sourceDataBase and
 * targetDataBase attribute. Class set all other attributes to the default values.
 * @param sourceDataBase defines the place where you're put the database tables.
 * @param targetDataBase defines the place where the target database tables are.
 * @param sourceType defines the type of the source database.
 * @param targetType defines the type of the target database.
 * @throws LoaderException
 */
public LoaderGenerator(String sourceDataBase, String targetDataBase, String sourceType,
                       String targetType, String confJarStructure ) throws LoaderException {

  generatorParameters=new InputParameters("");
  generatorParameters.setConfJarStructure(confJarStructure);
  generatorParameters.setSourceDataBase(sourceDataBase);

  generatorParameters.setTargetDataBase(targetDataBase);
  generatorParameters.setSourceType(sourceType);
  generatorParameters.setTargetType(targetType);
  generatorParameters.setSourceDriverProperties();
  generatorParameters.setTargetDriverProperties();
  generatorParameters.setOidAndVersionColumnName();
}







  /**
   * Main method controls parameters, it they are OK starts load method, but if they aren't makes alert.
   * @param args represents the input parameters
   *
   * Usage: java org.webdocwf.util.loader.generator.Generator [options]
   * Options:
   * -st sourceType, defines the type of the source database.
   * -tt targetType, defines the type of the target database.
   * -sdn sourceDriverName, defines the driver of defined source database.
   * -tdn targetDriverName, defines the driver of defined  target database.
   * -sdb sourceDataBase, defines the place where you're put the source database tables.
   * -tdb targetDataBase, defines the place where you're put the target database tables.
   * -m valueMode, defines the difference from overwrite and update attribute. If the valueMode
   * has value 'copy', attribute has value 'Overwrite', or if the valueMode has value 'sync', attribute
   * has value 'Update'. Default value of valueMode is 'copy'.
   * -o generatorOutput, is optional. That represent the directory, where the Octopus place created files.
   * If this argument don't exists, Octopus place created files in to current directory!
   * -su sourceUser, defines the user of the source database.
   * -sp sourcePassword, defines the user password for the source database.
   * -tu targetUser, defines the user of the target database.
   * -tp targetPassword, defines the user password for the target database.
   * -doml domlUrl, defines the place where the doml file is placed.
   * -pack packageName, defines the package name for the generated doml file.
   * -xml -xml generateXml, defines if you want to generate xml files as output files
   * -gdoml generateDoml, defines if you want to generate doml file as output file.
   * -sqlCT generateCreateTables defines if you want to generate only Sql statement for CREATE TABLE.
   * -sqlDT generateDropTableStmt defines if you want to generate only Sql statement for DROP TABLE.
   * -sqlCPK generateCreatePrimaryKeys defines if you want to generate only Sql statement for ALTER TABLE ADD CONSTRAINT (PRIMARY KEY).
   * -sqlCFK generateCreateForeigKeys defines if you want to generate only Sql statement for ALTER TABLE ADD CONSTRAINT (FOREIGN KEY).
   * -sqlCI generateCreateIndex defines if you want to generate only Sql statement for CREATE INDEX.
   * -sqlAll generateSqlForAllVendors defines if you want to generate sql statements for all named database vendors.
   * -fm fullMode defines the mode of output xml files. Possible values are true and false.
   */

  public static void main(String args[]) {

    String generatorOutput=null;
    String valueMode=null;
    String domlUrl=null;
    String packageName=null;

    String sourceType=null;
    String sourceDataBase=null;
    String sourceDriverName=null;
    String sourceUser=null;
    String sourcePassword=null;

    String targetType=null;
    String targetDataBase=null;
    String targetDriverName=null;
    String targetUser=null;
    String targetPassword=null;

    //Output mode
    //main
    String generateXml=null;
    String generateDoml=null;
    //for generate sql
    String generateDropTableStmt=null;
    String generateDropIntegrityStmt=null;
    String generateCreateTableStmt=null;
    String generateCreatePKStmt=null;
    String generateCreateFKStmt=null;
    String generateCreateIndexStmt=null;

    //for genrate sql for all vendors
    String generateSqlForAllVendors=null;
    //optimized modes
    String fullMode=null;
    String restoreMode=null;
    String includeTableList=null;
    String confJarStructure=null;

    if(args.length > 8 && args.length < 53){
      for (int i = 0; i < args.length - 1; i = i + 1) {
        if (args[i].equalsIgnoreCase("-st"))
          sourceType = args[++i];
        else if (args[i].equalsIgnoreCase("-sdb"))
          sourceDataBase = args[++i];
        else if (args[i].equalsIgnoreCase("-m"))
          valueMode = args[++i];
        else if (args[i].equalsIgnoreCase("-o"))
          generatorOutput = args[++i];
        else if (args[i].equalsIgnoreCase("-sdn"))
          sourceDriverName = args[++i];
        else if (args[i].equalsIgnoreCase("-tdn"))
          targetDriverName = args[++i];
        else if (args[i].equalsIgnoreCase("-tdb"))
          targetDataBase = args[++i];
        if (args[i].equalsIgnoreCase("-tt"))
          targetType = args[++i];
        if (args[i].equalsIgnoreCase("-su"))
          sourceUser = args[++i];
        if (args[i].equalsIgnoreCase("-sp"))
          sourcePassword = args[++i];
        if (args[i].equalsIgnoreCase("-tu"))
          targetUser = args[++i];
        if (args[i].equalsIgnoreCase("-tp"))
          targetPassword = args[++i];
        if (args[i].equalsIgnoreCase("-doml"))
          domlUrl = args[++i];
        if (args[i].equalsIgnoreCase("-pack"))
          packageName = args[++i];
        if (args[i].equalsIgnoreCase("-xml"))
          generateXml = args[++i];
        if (args[i].equalsIgnoreCase("-gdoml"))
          generateDoml = args[++i];
        if (args[i].equalsIgnoreCase("-sqlCT"))
          generateCreateTableStmt = args[++i];
        if (args[i].equalsIgnoreCase("-sqlDT"))
          generateDropTableStmt = args[++i];
        if (args[i].equalsIgnoreCase("-sqlDI"))
          generateDropIntegrityStmt = args[++i];
        if (args[i].equalsIgnoreCase("-sqlCPK"))
          generateCreatePKStmt = args[++i];
        if (args[i].equalsIgnoreCase("-sqlCFK"))
          generateCreateFKStmt = args[++i];
        if (args[i].equalsIgnoreCase("-sqlCI"))
          generateCreateIndexStmt = args[++i];
        if (args[i].equalsIgnoreCase("-sqlAll"))
          generateSqlForAllVendors = args[++i];
        if (args[i].equalsIgnoreCase("-fm"))
          fullMode = args[++i];
        if (args[i].equalsIgnoreCase("-rm"))
          restoreMode = args[++i];
        if (args[i].equalsIgnoreCase("-it"))
          includeTableList = args[++i];
        if (args[i].equalsIgnoreCase("-cjs"))
          confJarStructure = args[++i];
      }
    }
    else {
      System.out.println("You didn't enter all required parameters, for running the application!");
      BufferClass.getInstance().writeToBuffer("You didn't enter all required parameters, for running the application!");
      printUsage();
      System.exit(1);
    }
    try{

      LoaderGenerator generator=new LoaderGenerator(sourceType, sourceDataBase, valueMode,
          generatorOutput, sourceDriverName, targetDriverName, targetDataBase, targetType,
          sourceUser, sourcePassword, targetUser, targetPassword, domlUrl,
          packageName,generateDropTableStmt,generateDropIntegrityStmt,generateCreateTableStmt,generateCreatePKStmt,
          generateCreateFKStmt,generateCreateIndexStmt,generateSqlForAllVendors,
          generateXml, generateDoml, fullMode, restoreMode, includeTableList,
          confJarStructure );

      generator.generate();
      System.exit(0);

    } catch (LoaderException le) {
      System.out.println(le.getMessage());
      BufferClass.getInstance().writeToBuffer(le.getMessage());
      return;
    } catch (Throwable th) {
      BufferClass.getInstance().writeToBuffer(th.getMessage());
      th.printStackTrace(System.out);
      return;
    }
  }


  static void printUsage () {
    System.out.println("Usage: java org.webdocwf.util.loader.generator.Generator [options] ");
    BufferClass.getInstance().writeToBuffer("Usage: java org.webdocwf.util.loader.generator.Generator [options] ");
    System.out.println(" Options:");
    BufferClass.getInstance().writeToBuffer(" Options:");
    System.out.println(" -st sourceType, defines the type of the source database.");
    BufferClass.getInstance().writeToBuffer(" -st sourceType, defines the type of the source database.");
    System.out.println(" -tt targetType, defines the type of the target database.");
    BufferClass.getInstance().writeToBuffer(" -tt targetType, defines the type of the target database.");
    System.out.println(" -sdn sourceDriverName, represents driver of defined source database.");
    BufferClass.getInstance().writeToBuffer(" -sdn sourceDriverName, represents driver of defined source database.");
    System.out.println(" -tdn targetDriverName, represents driver of defined target database.");
    BufferClass.getInstance().writeToBuffer(" -tdn targetDriverName, represents driver of defined target database.");
    System.out.println(" -sdb sourceDatabase, defines the place where you're put the database tables");
    BufferClass.getInstance().writeToBuffer(" -sdb sourceDatabase, defines the place where you're put the database tables");
    System.out.println(" -tdb targetDataBase defines the place where the target database tables are.");
    BufferClass.getInstance().writeToBuffer(" -tdb targetDataBase defines the place where the target database tables are.");
    System.out.println(" -m valueMode, is the difference from overwrite and update attribute");
    BufferClass.getInstance().writeToBuffer(" -m valueMode, is the difference from overwrite and update attribute");
    System.out.println(" -su sourceUser, defines the user of the source database.");
    BufferClass.getInstance().writeToBuffer(" -su sourceUser, defines the user of the source database.");
    System.out.println(" -sp sourcePassword, defines user password for the source database.");
    BufferClass.getInstance().writeToBuffer(" -sp sourcePassword, defines user password for the source database.");
    System.out.println(" -tu targetUser, defines the user of the target database.");
    BufferClass.getInstance().writeToBuffer(" -tu targetUser, defines the user of the target database.");
    System.out.println(" -tp targetPassword, defines user password for the target database.");
    BufferClass.getInstance().writeToBuffer(" -tp targetPassword, defines user password for the target database.");
    System.out.println(" -doml domlUrl, defines the place where the doml file is placed.");
    BufferClass.getInstance().writeToBuffer(" -doml domlUrl, defines the place where the doml file is placed.");
    System.out.println(" -o generatorOutput, is optional. That represent the directory, where the Octopus place"+
                       " created files. If this argument dont exists, Octopus place created files in to current direcory!");
    BufferClass.getInstance().writeToBuffer(" -o generatorOutput, is optional. That represent the directory, where the Octopus place"+
                                            " created files. If this argument dont exists, Octopus place created files in to current direcory!");
    System.out.println(" -pack packageName, defines the package name for the generated doml file.");
    BufferClass.getInstance().writeToBuffer(" -pack packageName, defines the package name for the generated doml file.");

    System.out.println(" -xml generateXml, defines if you want to generate xml files as output files. "+
                       "    Possible values are true and false.");
    BufferClass.getInstance().writeToBuffer(" -xml generateXml, defines if you want to generate xml files as output files. "+
                       "    Possible values are true and false.");
    System.out.println(" -gdoml generateDoml, defines if you want to generate doml file as output file. "+
                       "    Possible values are true and false.");
    BufferClass.getInstance().writeToBuffer(" -gdoml generateDoml, defines if you want to generate doml file as output file. "+
                       "    Possible values are true and false.");
    System.out.println(" -sqlCT generateCreateTables, defines if you want to generate only Sql statement for CREATE TABLE.");
    BufferClass.getInstance().writeToBuffer(" -sqlCT generateCreateTables, defines if you want to generate only Sql statement for CREATE TABLE.");

    System.out.println(" -sqlDT generateDropTableStmt, defines if you want to generate only Sql statement for DROP TABLE.");
    BufferClass.getInstance().writeToBuffer(" -sqlDT generateDropTableStmt, defines if you want to generate only Sql statement for DROP TABLE.");

    System.out.println(" -sqlCPK generateCreatePrimaryKeys, defines if you want to generate only Sql statement for ALTER TABLE ADD CONSTRAINT (PRIMARY KEY).");
    BufferClass.getInstance().writeToBuffer(" -sqlCPK generateCreatePrimaryKeys, defines if you want to generate only Sql statement for ALTER TABLE ADD CONSTRAINT (PRIMARY KEY).");

    System.out.println(" -sqlCFK generateCreateForeigKeys, defines if you want to generate only Sql statement for ALTER TABLE ADD CONSTRAINT (FOREIGN KEY).");
    BufferClass.getInstance().writeToBuffer(" -sqlCFK generateCreateForeigKeys, defines if you want to generate only Sql statement for ALTER TABLE ADD CONSTRAINT (FOREIGN KEY).");

    System.out.println(" -sqlCI generateCreateIndex, defines if you want to generate only Sql statement for CREATE INDEX.");
    BufferClass.getInstance().writeToBuffer(" -sqlCI generateCreateIndex, defines if you want to generate only Sql statement for CREATE INDEX.");

    System.out.println(" -sqlAll generateSqlForAllVendors, defines if you want to generate sql statements for all named database vendors.");
    BufferClass.getInstance().writeToBuffer(" -sqlAll generateSqlForAllVendors, defines if you want to generate sql statements for all named database vendors.");

    System.out.println(" -fm fullMode, defines the mode of output xml files. Possible values are true and false.");
    BufferClass.getInstance().writeToBuffer("  -fm fullMode, defines the mode of output xml files. Possible values are true and false.");

    System.out.println(" -it includeTableList, defines tables which you want to incude into Generator process.");
    BufferClass.getInstance().writeToBuffer("  -it includeTableList, defines tables which you want to incude into Generator process.");

    System.out.println("For more details, see documentation.");
    BufferClass.getInstance().writeToBuffer("For more details, see documentation.");

  }

  /**
   *
   * Method generate is main method in class LoaderGenerator. It is used to load data from the source
   * table, and starts LoadJobWriter,CreateIncludeFiles and CreateIncludeDomlFiles methods.
   * @throws LoaderException
   */
  public void generate() throws LoaderException {
    try{
      LoadJobWriter loadJobWriter;
      GenerateSqlForAllVendors generateSqlForAllVendors;
      if(generatorParameters.getSqlForAllVendors()==true){
        generateSqlForAllVendors=new GenerateSqlForAllVendors(generatorParameters);
      }else{
        if(generatorParameters.getDomlPath() == null ||
           generatorParameters.getDomlPath().equalsIgnoreCase("")) {
          CreateIncludeFiles createIncludeFiles=new CreateIncludeFiles(generatorParameters);
        } else {
          CreateIncludeDomlFiles createIncludeDomlFiles;
          createIncludeDomlFiles=new CreateIncludeDomlFiles(generatorParameters);
        }
        if(generatorParameters.getGenerateXml().equalsIgnoreCase("true"))
          loadJobWriter= new LoadJobWriter(generatorParameters);
      }
    }catch(Exception e){
      String msg="Exception in method generate() in LoaderGenerator class!";
      LoaderException le = new LoaderException(msg+ "\n" +e.getMessage(), (Throwable)e);
      throw le;

    }
  }
}